class WaitResponse:
    def __init__(self):
        self.packet = None

    def create_promise(self):
        while self.packet is None:  # Wait for condition to be true (Not very elegant)
            pass
        return self.packet

    def receive_packet(self, packet):
        self.packet = packet



from pyfabric.Entity.Inventory import Inventory
from pyfabric.Networking.Client import Client


class Location:
    def __init__(self, uuid: str, client: Client):
        self.client = client
        self.uuid = uuid
        self.x, self.y, self.z, self.yaw, self.pitch = 0, 0, 0, 0, 0
        self.update()

    def update(self):
        raw_location = self.client.request("get_entity_location", self.uuid)
        self.x, self.y, self.z, self.yaw, self.pitch = raw_location["data"]

    def __str__(self):
        return "location:{x:~%d,y:~%d,z:~%d,yaw:~%d,pitch:~%d}" % (
            self.x, self.y, self.z, self.yaw, self.pitch)


class Entity:
    def __init__(self, client: Client, uuid: str):
        self.uuid = uuid
        self.client = client
        self.location = Location(uuid, client)

    def __str__(self):
        return "{entity:{uuid:%s,location:%s}}" % (self.uuid, self.location)


class Player(Entity):
    def __init__(self, client: Client, uuid: str):
        super(Player, self).__init__(client, uuid)
        self.client = client

    def get_block_looking_at(self):
        return self.client.request_string("block_looking_at", 0, self.uuid)

    def get_dimension(self):
        return self.client.request_string("get_client_player_world", 0)

    def send_chat(self, message: str):
        """
        Send a message to the player

        Note: messages in minecraft can only be 256 characters long (Socket transfer limit)
        :param message: The message the player should receive. Use § for colors
        """
        self.client.notify_server("client_chat", message)

    def send_actionbar(self, action_bar):
        self.client.notify_server("client_actionbar", action_bar)

    def get_inventory(self) -> Inventory:
        return Inventory(self.client, self.uuid)



import time

from pyfabric.Networking.Client import Client


class ClientWorld:
    def __init__(self, client: Client):
        self.client = client

    def get_entities(self):
        """
        This will return a generator with all entites in the world
        Note: It
        :return: A generator function for all entities
        """
        self.client.notify_server("reset_entities")
        time.sleep(0.1)  # make sure the socket flushed before sending the other requests
        raw_uuid = self.client.request_string("get_all_entities_in_world_single", 0)
        while raw_uuid != "CONSUMED":
            yield raw_uuid
            raw_uuid = self.client.request_string("get_all_entities_in_world_single", 0)

    def spawn_particle(self):
        """
        Has no implementation yet - So don't use it
        :return:
        """
        pass

    def play_sound(self):
        """
        Has no implementation yet - So don't use it
        :return:
        """
        pass

    def get_players(self):
        self.client.notify_server("reset_players")
        time.sleep(0.1)  # make sure the socket flushed before sending the other requests
        raw_uuid = self.client.request_string("get_all_players_in_world_single", 0)
        while raw_uuid != "CONSUMED":
            yield raw_uuid
            raw_uuid = self.client.request_string("get_all_players_in_world_single", 0)

    def get_block_at(self, x: int, y: int, z: int):
        material = self.client.request_string("get_block_at", 0, x, y, z)
        return material



import json
import socket
import threading
import time
import uuid

from pyfabric.Networking.Response import WaitResponse


class ResponseError(Exception):
    def __init__(self, text: str):
        super(ResponseError, self).__init__(text)


class Client:
    def __init__(self, handle_event, port: int = 1337):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(1)

        try:
            self.socket.connect(("localhost", port))
        except (ConnectionRefusedError, socket.timeout):
            print("[Client]: Couldn't connect with the fabric mod")
            self.connected = False
            exit(0)
        else:
            print("[Client]: Connect with the pyfabric mod")

        self.handle_event = handle_event
        self.waiting_response = {}
        self.uuid = uuid.uuid4()

        self.connected = True
        self.worker = threading.Thread(target=self.__request_worker__)
        self.worker.start()

    def __request_worker__(self):
        # Start listening to server response
        while self.connected:
            try:
                raw = self.socket.recv(1024).decode('utf-8')
                if not raw:
                    continue

                msg = json.loads(raw)
                if not self.connected:
                    continue

                if "id" not in msg:
                    print(f"[WARN]: Received an event with no id. Event response: {msg}")
                    continue

                msg_id = msg["id"]
                if msg_id.startswith("__"):
                    print(msg)
                elif msg_id in self.waiting_response:
                    self.waiting_response.get(msg_id)(msg)
                    del self.waiting_response[msg_id]
                else:
                    emit_thread = threading.Thread(target=self.execute_event, args=[msg_id, msg])
                    emit_thread.start()

            except socket.timeout:  # Ignored
                pass
            except ConnectionResetError:
                print("[Client]: Server closed. Stopping services")
                self.connected = False
                exit(0)
                return

    def execute_event(self, event_id, msg):
        self.handle_event(event_id, msg)

    def request_float(self, registry: str, index: int, *data) -> float:
        return self.request_item(registry, index, *data)

    def request_int(self, registry: str, index: int, *data) -> int:
        return self.request_item(registry, index, *data)

    def request_string(self, registry: str, index: int, *data) -> str:
        return self.request_item(registry, index, *data)

    def request_item(self, registry: str, index: int, *data):
        response = self.request(registry, *data)
        if "data" not in response:
            raise ResponseError(f"Response doesn't contain a data tag. Response: {response}")
        data_array = response.get("data")
        if index >= len(data_array):
            raise ResponseError(f"Response index out of index. Response: {response}")
        return data_array[index]

    def request(self, registry: str, *data) -> dict:
        package = {
            "id": registry,
            "data": data
        }

        return self.request_raw(package)

    def notify_server(self, registry: str, *data) -> None:
        if not self.connected:
            return

        packet = {
            "id": "no_response_" + registry,
            "data": data
        }

        self.socket.send(bytes(json.dumps(packet).encode('utf-8')))
        time.sleep(0.1)  # Make sure everything got flushed

    def request_raw(self, packet: dict) -> dict:
        if not self.connected:
            return {}

        response = WaitResponse()
        self.waiting_response[packet["id"]] = response.receive_packet

        self.socket.send(bytes(json.dumps(packet).encode('utf-8')))

        # Get server response
        return response.create_promise()



import time

from pyfabric.Networking.Client import Client


class ItemStack:
    def __init__(self,
                 registry: str,
                 count: int,
                 max_count: int,
                 damage: int,
                 max_damage: int,
                 name: str,
                 rarity: str,
                 repair_cost: int,
                 is_enchanted: bool,
                 is_damageable: bool,
                 is_food: bool,
                 nbt: str):
        self.registry = registry

        self.count = count
        self.max_count = max_count

        self.damage = damage
        self.max_damage = max_damage

        self.name = name
        self.rarity = rarity
        self.repair_cost = repair_cost
        self.is_enchanted = is_enchanted
        self.is_damageable = is_damageable
        self.is_food = is_food
        self.nbt = nbt

    def __str__(self):
        return "{item:{registry:%s,count:%d,max_count:%d,damage:%d," \
               "max_damage%d,name:%s,rarity:%s,repair_cost:%d," \
               "is_enchanted:%s,is_damageable:%s,is_food:%s, nbt:%s}}" % \
               (self.registry, self.count, self.max_count, self.damage, self.max_damage,
                self.name, self.rarity, self.repair_cost,
                str(self.is_enchanted), str(self.is_damageable), str(self.is_food), self.nbt)


class Inventory:
    def __init__(self, client: Client, entity_uuid: str):
        self.entity_uuid = entity_uuid
        self.client = client

    def get_mainhand(self) -> ItemStack:
        items = self.client.request("get_entity_inventory_main_hand", self.entity_uuid)
        if "data" not in items:
            return ItemStack("minecraft:air", 0, 64, 0, 0, "translation{key='item.minecraft.air', args=[]}", "COMMON",
                             0, False, False, False, "{}")
        return ItemStack(*items["data"])

    def get_offhand(self) -> ItemStack:
        items = self.client.request("get_entity_inventory_off_hand", self.entity_uuid)
        return ItemStack(*items["data"])

    def get_armor(self):
        self.client.notify_server("get_entity_inventory_armor:reset")
        time.sleep(0.1)  # make sure the socket flushed before sending the other requests
        raw_stack = self.client.request("get_entity_inventory_armor:iterator", self.entity_uuid)
        while "data" in raw_stack and raw_stack["data"][0] != 'CONSUMED':
            if raw_stack["data"] != "NULL":
                yield ItemStack(*raw_stack["data"])

            raw_stack = self.client.request("get_entity_inventory_armor:iterator", self.entity_uuid)

    def get_selected_slot(self) -> int:
        return self.client.request_int("get_entity_inventory_selected_slot", 0, self.entity_uuid)

    def get_item_in_slot(self, slot: int) -> ItemStack:
        items = self.client.request("get_entity_inventory_stack_at", slot)
        if "data" not in items or items["data"][0] == "NULL":
            return ItemStack("minecraft:air", 0, 64, 0, 0, "translation{key='item.minecraft.air', args=[]}", "COMMON",
                             0, False, False, False, "{}")

        return ItemStack(*items["data"])






from enum import Enum, auto


class ArgumentType(Enum):

    def _generate_next_value_(self, start, count, last_values):
        return self

    SUB = auto()
    BOOL = auto()
    FLOAT = auto()
    DOUBLE = auto()
    INTEGER = auto()
    LONG = auto()
    STRING = auto()
    GREEDY_STRING = auto()
    BLOCK_POS = auto()
    VEC3 = auto()
    VEC2 = auto()
    COLOR = auto()
    COMPONENT = auto()
    MESSAGE = auto()
    NBT_COMPOUND_TAG = auto()
    NBT_TAG = auto()
    OPERATION = auto()
    PARTICLE = auto()
    ANGLE = auto()
    ROTATION = auto()
    ITEM_SLOT = auto()
    DIMENSION = auto()
    TIME = auto()
    UUID = auto()
    INT_RANGE = auto()
    FLOAT_RANGE = auto()


class CommandBuilder:
    def __init__(self):
        self.__literal__ = ""
        self.__argument_type__ = ArgumentType.SUB
        self.__subs__ = []

    def set_name(self, name: str):
        self.__literal__ = name
        return self

    def set_arg(self, arg: ArgumentType):
        self.__argument_type__ = arg
        return self

    def add_sub_command(self, cmd: str):
        self.__subs__.append(cmd)
        return self

    def build(self) -> str:
        subs = ", ".join(self.__subs__)

        return '{"literal":"%s","type":"%s","subCommands": [%s]}' % (self.__literal__, self.__argument_type__.name, subs)